/*
    cyberlinkusb -- catches signals from CyberLink remote control and simulates  
      key presses.
    Copyright (C) 2008  Thaler <thaj at users.sourceforge.net>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <usb.h>
#include <X11/Xlib.h>
#include <X11/extensions/XTest.h>
#include <X11/keysym.h>
#include <pthread.h>
#include <signal.h>

#include "keys.h"
#include "config.h"

#define VERSION 20081207
#define USB_VENDOR 0x766
#define USB_PRODUCT 0x204

int terminated = False;
int verbose = False;
char *display_name = NULL;

struct watchInterfaceArg {
  usb_dev_handle *handle;
  struct usb_device *dev;
  int interface;
};

int RET1 = 1;
int RET0 = 0;

int findKey(char *buf, int len) {
  if(len > 8) return 0;

  int i, u;
  int b;
  for(i = 0; lookupKeyTable[i].key != KEY_NONE; i++) {
    b = 0;
    
    for(u = 0; u < len && lookupKeyTable[i].bytes[u] != -1; u++) {
      if(lookupKeyTable[i].bytes[u] == buf[u]) {
        b++;
      } else {
        break;
      }
    }
    
    if(b == len && lookupKeyTable[i].bytes[b] == -1) {
      //printf("Key: %i\n", lookupKeyTable[i].key);
      return KeyToXK[lookupKeyTable[i].key];
    }
  }

  return 0;
}

void SendKey(Display *display, int keycode) {
  XTestFakeKeyEvent(display, keycode, True, CurrentTime);
  XTestFakeKeyEvent(display, keycode, False, CurrentTime);
  XSync(display, True);
}

#define USBREADERROR fprintf(stderr, "Interface %i: Read Error: %i, %s\n", a->interface, r, usb_strerror()); \
  ret = &RET1; \
  break;          

#define RETURN1 usb_close(a->handle);return &RET1;

#define USB_READ usb_interrupt_read
//#define USB_READ usb_bulk_read

void *watchInterface(void *arg) {
  struct watchInterfaceArg *a = ((struct watchInterfaceArg *)arg);
  
  a->handle = usb_open(a->dev);
  
  if(a->handle == NULL) {
    fprintf(stderr, "Interface %i: Unable to open USB device: %s\n", a->interface, usb_strerror());
    return &RET1;
  }  
  
  // XXX: Hard-coded values for bInterfaceNumber and bEndpointAddress ?

  if(a->dev->config[0].interface[a->interface].num_altsetting == 0) {
    fprintf(stderr, "Interface %i: No settings found.\n", a->interface);  
    RETURN1
  }
  
  // get bInterfaceNumber
  u_int8_t bInterfaceNumber = a->dev->config[0].interface[a->interface].altsetting[0].bInterfaceNumber;

  if(a->dev->config[0].interface[a->interface].altsetting[0].bNumEndpoints == 0) {
    fprintf(stderr, "Interface %i: No Endpoints found.\n", a->interface);  
    RETURN1
  }

  // get bEndpointAddress
  u_int8_t bEndpointAddress = a->dev->config[0].interface[a->interface].altsetting[0].endpoint[0].bEndpointAddress;

  
  char driver[20] = "";
  if(usb_get_driver_np(a->handle, bInterfaceNumber, driver, 20) < 0) {
    fprintf(stderr, "Interface %i: usb_get_driver_np: %s\n", a->interface, usb_strerror());
  } else {
    fprintf(stderr, "Interface %i: Driver name: %s\n", a->interface, driver);
  }
        
  if(usb_detach_kernel_driver_np(a->handle, bInterfaceNumber) < 0) {
    fprintf(stderr, "Interface %i: usb_detach_kernel_driver_np: %s\n", a->interface, usb_strerror());
  }
  
  if(usb_claim_interface(a->handle, bInterfaceNumber) < 0) {
    fprintf(stderr, "Interface %i: Unable to claim interface: %s\n", a->interface, usb_strerror());
    RETURN1
  }
  
  usb_set_altinterface(a->handle, 0);  
  
  // open X display
  Display *display = NULL;
  display = XOpenDisplay(display_name); 
  if(display == NULL) {
    fprintf(stderr, "Interface %i: Cannot open X display.\n", a->interface);
    usb_release_interface(a->handle, bInterfaceNumber);
    RETURN1
  }    
  
  //if(usb_clear_halt(a->handle, bEndpointAddress) <0) printf("error\n");
  
  char buf[8];
  int r;
  int *ret = &RET0;

  // clear the buffer
  // i have no better idea
  while(USB_READ(a->handle, bEndpointAddress, buf, 8, 500) > 0) {}

  int keysym, keycode;

  while(!terminated) {
    if((r = USB_READ(a->handle, bEndpointAddress, buf, 8, 30000)) > 0) {
        
        if(verbose) {
          fprintf(stderr, "Interface %i: %i bytes read: ", a->interface, r);
            for(int u = 0; u < r; u++)
              fprintf(stderr, "0x%x ", buf[u]);
            fprintf(stderr, "\n");
        }
        
        // find key
        keysym = findKey(buf, r);
        
        if(verbose) {
          if(keysym == 0) {
            fprintf(stderr, "Interface %i: No Key found.\n", a->interface);  
          } else {
            fprintf(stderr, "Interface %i: Key found: %i (%s)\n", a->interface, keysym, XKeysymToString(keysym));
          }
        }

        if(keysym != 0) {
          keycode = XKeysymToKeycode(display, keysym);
          
          SendKey(display, keycode);
          
          /* Key Repeat */
          
          // first key delay
          r = USB_READ(a->handle, bEndpointAddress, buf, 8, (int)(KEY_DELAY1));
          if(r == -ETIMEDOUT) {
            SendKey(display, keycode);
            
            while(!terminated && (r = USB_READ(a->handle, bEndpointAddress, buf, 8, (int)(KEY_DELAY2))) == -ETIMEDOUT) {
              SendKey(display, keycode);
            }
            
            if(r < 0 && r != -ETIMEDOUT) {
              USBREADERROR
            }
          } else if (r < 0) {
            USBREADERROR            
          }
        }
      } else {
        if(r != -ETIMEDOUT)  {
          USBREADERROR
        }
      }
  }
    
  usb_release_interface(a->handle, bInterfaceNumber);
  
  XCloseDisplay(display);
  
  usb_close(a->handle);
  
  return ret;
}

int runthreads(struct usb_device *dev) {
  pthread_t interface1, interface2;  
  struct watchInterfaceArg a1, a2;
  
  a1.handle = NULL;
  a1.dev = dev;
  a1.interface = 0;
  
  a2 = a1;
  a2.interface = 1;
  
  pthread_create(&interface1, NULL, watchInterface, &a1);
  pthread_create(&interface2, NULL, watchInterface, &a2);
  
  void *v1 = NULL, *v2 = NULL;

  pthread_join(interface1, &v1);
  pthread_join(interface2, &v2);
  
  // safe is safe
  return (v1 == NULL || v2 == NULL) ? 1 : *((int *)v1) | *((int *)v2);
}

struct usb_device *lookup() {
  struct usb_bus *bus;
  struct usb_device *dev;
  
  for(bus = usb_get_busses(); bus; bus = bus->next) {
    if(verbose) fprintf(stderr, "Bus: %s\n", bus->dirname);

    for(dev = bus->devices; dev; dev = dev->next) {
      if(verbose) {
        fprintf(stderr, "  Device: %s\n", dev->filename);
        fprintf(stderr, "    idVendor: %x\n", dev->descriptor.idVendor);
        fprintf(stderr, "    idProduct: %x\n", dev->descriptor.idProduct);
      }
      
      if(USB_VENDOR == dev->descriptor.idVendor && USB_PRODUCT == dev->descriptor.idProduct) {
        if(verbose) fprintf(stderr, " ==> Found\n");
        return dev;
      }
    }
  }
  
  return NULL;
}

void signal_handler(int sig) {
  printf("Please wait (max. 30 sec) ...\n");
  terminated = True;
}

void usage(char *name) {
  printf("Usage: %s [-d display] [-v] [-l level] | -h \n\
Catches with libusb signals from CyberLink remote control and simulates key presses in X.\n\n\
  -d display  set X display\n\
  -l level    set debug level for libusb (default 0=disabled)\n\
  -v          verbose mode\n\
  -V          display the version\n\
  -h          print this text\n\
\n\
To change the keys, you have to recompile the source code and do not forget to change the keys.\n\
\n\
Program written by Thaler (thaj <at> users.sourceforge.net).\n\n\
Feel free to vist:\n\
  http://linux.thaj.net63.net/cyberlinkusb/\n\
  http://jtspainwithlinux.blogspot.com/\n\
  http://onnetblog.atspace.org/\n\n", name);
}

int main(int argc, char *argv[]) {
  int usb_debug_level = 0;
    
  /* Parse command line options.  */
  int c;
  while ((c = getopt(argc, argv, "d:vl:hV")) != -1) {
    switch (c) {
      case 0:
        break;
      case 'd':
        display_name = optarg;
        break;
      case 'v':
        verbose = True;
        break;
      case 'h':
        usage(argv[0]);
        return 0;
      case 'V':
        printf("version %i\n", VERSION);
        return 0;
      case 'l':
        usb_debug_level = atoi(optarg);
        break;
      default: 
        usage(argv[0]);
        return 1;
    }
  }
  
  if(usb_debug_level > 0)
    usb_set_debug(usb_debug_level);

  /* init usb */
  usb_init();
  
  if(usb_find_busses() <= 0) {
    fprintf(stderr, "Unable to find USB busses\n");
    return 1;
  }
  
  if(usb_find_devices() <= 0) {
    fprintf(stderr, "Unable to find USB devices\n");
    return 1;
  }
  
  // searching for the cyberlink usb device
  struct usb_device *dev = lookup();
  
  if(dev == NULL) {
    fprintf(stderr, "Unable to find CyberLink USB device.\nUse -v and -l for verbose output.\n");
    return 1;
  }
  
  if(dev->descriptor.bNumConfigurations != 1) {
    fprintf(stderr, "Expected 1 configuration but %i found.\n", dev->descriptor.bNumConfigurations);  
    return 1;
  }
  
  if(dev->config[0].bNumInterfaces != 2) {
    fprintf(stderr, "Expected 2 interfaces but %i found.\n", dev->config[0].bNumInterfaces);  
    return 1;
  }
  
  signal(SIGQUIT, signal_handler);
  signal(SIGTERM, signal_handler);
  signal(SIGINT, signal_handler);
  
  int r = runthreads(dev);
  
  return r;
}
